#!/usr/bin/env python
# coding: utf-8
"""Test ezq functions."""

# native
from functools import partial
import time

# pkg
import ezq


def msg_counter(q: ezq.Queue, n_msg: int, num: int = 0):
    """Count up the number of messages.

    Args:
        q (ezq.Queue): queue to read from
        n_msg (int): expected number of messages
        num (int, optional): process number. Defaults to 0.
    """
    assert num <= ezq.NUM_CPUS, "expect subprocess number to be less than cpus"

    count = sum(msg.data for msg in ezq.iter_msg(q))
    assert count == n_msg, f"expect {n_msg} messages"


def msg_summer(q: ezq.Queue, n_msg: int, n_sum: int):
    """Add up the message content.

    Args:
        q (ezq.Queue): queue to read from
        n_msg (int): expected number of messages
        n_sum (int): expected sum of message data
    """

    result, count = 0, 0
    for msg in ezq.iter_msg(q):
        result += msg.data
        count += 1
    assert count == n_msg, f"expect {n_msg} messages"
    assert result == n_sum, f"expect sum of {n_sum}"


def test_all_msg():
    """Expect iterator to iterate over the queue."""
    num = 1000
    q = ezq.Queue()
    for _ in range(num):
        q.put(ezq.Msg("MSG", 1))
    q.put(ezq.END_MSG)

    total = 0
    for msg in ezq.iter_msg(q):
        total += msg.data
    assert total == num, "expect iterator to get all messages"


def test_end_early():
    """Expect iterator to end early."""
    num = 1000
    num_stop = 900  # give the queue time to warm up
    q = ezq.Queue()
    for _ in range(num):
        q.put(ezq.Msg("MSG", 1))
    q.put(ezq.END_MSG)

    total = 0
    for msg in ezq.iter_msg(q):
        total += msg.data
        if total == num_stop:
            ezq.stop_iter_msg()
    assert total == num_stop, "expect iterator to stop early"


def test_single():
    """Test single subprocess."""
    n_msg = 1000
    n_sum = sum(range(n_msg))

    q = ezq.Queue()
    workers = ezq.start(partial(msg_summer, q, n_msg, n_sum), 1)

    for i in range(n_msg):
        q.put(ezq.Msg(data=i))

    ezq.wait(q, workers)
    time.sleep(0.1)  # give it time to finish


def test_numbered():
    """Test numbered processes."""
    n_msg = 1000
    q = ezq.Queue()
    workers = ezq.start_numbered(partial(msg_counter, q))

    for _ in range(n_msg):
        q.put(ezq.Msg(data=1))

    ezq.wait(q, workers)
    time.sleep(0.1)  # give it time to finish



#!/usr/bin/env python
# coding: utf-8
"""Simple wrapper for python multiprocessing."""

# native
from dataclasses import dataclass
from functools import partial
from multiprocessing import Process, Queue
from os import cpu_count
from typing import Any, Callable, List, Iterator
import queue
import signal

# pkg
from .__about__ import (
    __url__,
    __version__,
    __pubdate__,
    __author__,
    __email__,
    __copyright__,
    __license__,
)


@dataclass
class Msg:
    """Message for a queue."""

    kind: str = ""
    data: Any = None


END_MSG = Msg(kind="END")
"""Message that indicates no future messages will be sent."""

IS_ALIVE = True
"""Whether the process is alive. NOTE: Setting to True will prevent `queuer` from working."""

NUM_CPUS = cpu_count() or 1
"""Number of CPUs on this machine."""

Daemon = lambda target: Process(daemon=True, target=target)


def stop_iter_msg():
    """Stop `iter_msg` by setting `IS_ALIVE` to `False`."""
    global IS_ALIVE  # pylint: disable=global-statement
    IS_ALIVE = False


def iter_msg(q: Queue, timeout=0.05) -> Iterator[Msg]:
    """Yield messages for a queue.

    Note, this function checks whether the global `IS_ALIVE` value is `True`. If it
    is set to `False`, the queuer will immediately stop.

    Args:
        q (Queue): queue to read from
        timeout (float, optional): time in seconds to poll the queue. Defaults to 0.05.

    Returns:
        None

    Yields:
        Iterator[Msg]: message in the queue
    """
    while IS_ALIVE:
        try:
            msg = q.get(block=True, timeout=timeout)
            if msg.kind == END_MSG.kind:
                break
            yield msg
        except queue.Empty:  # pragma: no cover
            continue


def start_processes(procs: List[Process]) -> List[Process]:
    """Return a list of subprocesses after starting them.

    Args:
        procs (List[Process]): list of subprocesses to start

    Returns:
        List[Process]: subprocesses that were started
    """
    for proc in procs:
        proc.start()
    return procs


def start(worker: Callable, count: int = NUM_CPUS) -> List[Process]:
    """Start a worker as several subprocesses.

    Args:
        worker (Callable): function to run in each subprocess
        count (int, optional): number of subprocesses to start. Defaults to NUM_CPUS.

    Returns:
        List[Process]: subprocesses that were started
    """
    return start_processes([Daemon(worker) for _ in range(count)])


def start_numbered(worker: Callable, count: int = NUM_CPUS) -> List[Process]:
    """Start a worker and pass it its number.

    This is useful, for example, when using `tqdm` and you want to set the position.

    Args:
        worker (Callable): function to run in each subprocess
            (NOTE: must take a `num` parameter of type `int`)
        count (int, optional): number of subprocesses to start. Defaults to NUM_CPUS.

    Returns:
        List[Process]: subprocesses that were started
    """
    return start_processes([Daemon(partial(worker, num=n)) for n in range(count)])


def wait(q: Queue, procs: List[Process]):
    """Notify a list of processes to end and wait for them to join.

    Args:
        q (Queue): queue to put the end message on
        procs (List[Process]): processes to wait for
    """
    for _ in range(len(procs)):
        q.put(END_MSG)
    # all processes notified

    for proc in procs:
        proc.join()
    # all processes ended


signal.signal(signal.SIGINT, stop_iter_msg)



#!/usr/bin/env python
# coding: utf-8
"""Simple wrapper for python multiprocessing."""

__url__ = "https://github.com/metaist/ezq"
__version__ = "1.0.0"
__pubdate__ = "unpublished"  # date/time in UTC
__author__ = "Metaist LLC"
__email__ = "metaist@metaist.com"
__copyright__ = "Copyright 2021 by Metaist LLC"
__license__ = "MIT"



#!/usr/bin/env python3

"""Install library package."""

# native
from pathlib import Path
import site
import sys

# lib
from setuptools import setup, find_namespace_packages

# pkg
pkg = {}
here = Path(__file__).parent.resolve()
top = here / "src" / "ezq"
exec((top / "__about__.py").open().read(), pkg)  # pylint: disable=exec-used

# See: https://github.com/pypa/pip/issues/7953
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# See: https://github.com/pypa/pipenv/issues/1911
# See: https://caremad.io/posts/2013/07/setup-vs-requirement/

setup(
    python_requires=">=3.8",
    name="ezq",
    version=pkg["__version__"],
    description=pkg["__doc__"].split("\n")[0],
    long_description=pkg["__doc__"],
    license=pkg["__license__"],
    author=pkg["__author__"],
    author_email=pkg["__email__"],
    url=pkg["__url__"],
    download_url=pkg["__url__"],
    package_dir={"": "src"},
    packages=find_namespace_packages(where="src"),
    install_requires=["setuptools"],
    keywords=["simple", "multiprocessing", "queue", "subprocesses"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
    ],
)
